-- thomson.lua : lots of utility for handling
-- thomson TO8 screen.
--
-- Version 1.1
-- 
-- Copyright 2016 Samuel Devulder
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; version 2
-- of the License. See <http://www.gnu.org/licenses/>

if not thomson then

run("pixel.lua") -- optionnal

thomson = {}

-- RAM banks
thomson.ramA = {}
thomson.ramB = {}
for i=1,8000 do
	thomson.ramA[i] = 0
	thomson.ramB[i] = 0;
end

-- color levels
thomson.levels = {
	-- in pc-space (0-255):
	pc = {0,100,127,142,163,179,191,203,215,223,231,239,
		  243,247,251,255}, 
	-- in linear space (0-255):
	linear = {},
	-- maps pc-levels (0-255) to thomson levels (0-15)
	pc2to={}
};

-- pc space to linear space
local function toLinear(val)
	-- use the version from pixel library
	if not Pixel then
		val = val/255
		if val<=0.081 then 
			val = val/4.5; 
		else 
			val = ((val+0.099)/1.099)^2.2; 
		end
		val = val*255
		return val;
	else
		return Pixel:new(val,0,0):toLinear().r
	end
end
		
for i=1,16 do 
	thomson.levels.linear[i] = toLinear(thomson.levels.pc[i])
end
for i=0,255 do 
	local r,cm,dm = toLinear(i),0,1e30;
	for c,v in ipairs(thomson.levels.linear) do
		local d = math.abs(v-r);
		if d<dm then cm,dm = c,d; end
	end
	thomson.levels.pc2to[i] = cm;
end

-- palette stuff
function thomson.palette(i, pal) 
	-- returns palette #i if pal is missing (nil)
	-- if pal is a number, sets palette #i
	-- if pal is an array, sets the palette #i, #i+1, ...
	if type(pal)=='table' then
		for j,v in ipairs(pal) do
			thomson.palette(i+j-1,v)
		end
	elseif pal and i>=0 and i<thomson._palette.max then
		thomson._palette[i+1] = pal
	elseif not pal and i>=0 and i<thomson._palette.max then
		return thomson._palette[i+1]
	end
end;
thomson._palette = {offset = 0, max=16}
thomson.default_palette = {0,15,240,255,3840,3855,4080,4095,
						   1911,826,931,938,2611,2618,3815,123}

-- border color
function thomson.border(c)
	if c then
		thomson._border = c;
	else
		return thomson._border
	end
end
thomson.border(0)

-- helper to appen tables to tables
function thomson._append(result, ...)
	for _,tab in ipairs({...}) do
		for _,v in ipairs(tab) do
			table.insert(result,v)
		end
	end
end

-- RLE compression of data into result
function thomson._compress(result,data)
	local partial,p,pmax={},1,#data
	local function addCarToPartial(car)
		partial[2] = partial[2]+1
		partial[2+partial[2]] = car
	end
	while p<=pmax do
		local num,car = 1,data[p]
		while num<255 and p<pmax and data[p+1]==car do
			num,p = num+1,p+1
		end
		local default=true
		if partial[1] then
			-- 01 aa 01 bb ==> 00 02 aa bb
			if default and num==1 and partial[1]==1 then
				partial = {0,2,partial[2],car}
				default = false
			end
			-- 00 n xx xx xx 01 bb ==> 00 n+1 xx xx xx bb
			if default and num==1 and partial[1]==0 and partial[2]<255 then
				addCarToPartial(car)
				default = false
			end
			-- 00 n xx xx xx 02 bb ==> 00 n+2 xx xx xx bb bb (pas utile mais sert quand combin  la regle ci-dessus)
			if default and num==2 and partial[1]==0 and partial[2]<254 then
				addCarToPartial(car)
				addCarToPartial(car)
				default = false
			end
		end
		if default then
			thomson._append(result, partial)
			partial = {num,car}
		end
		p=p+1
	end
	thomson._append(result, partial)
	return result
end

-- saves the thomson screen into a MAP file
function thomson.savep(name)
	local data = thomson._get_map_data()
	local tmp = {0, math.floor(#data/256), #data%256,0,0}
	thomson._append(tmp,data,{255,0,0,0,0})
	local out = io.open(name,"wb")
	out:write(string.char(unpack(tmp)))
	out:close()
end

-- copy ramA/B onto GrafX2 screen
function thomson.updatescreen()
	-- refresh palette
	for i=1,thomson._palette.max do
		local v=thomson._palette[i]
		local r=v % 16
		local g=math.floor(v/16)  % 16
		local b=math.floor(v/256) % 16
		setcolor(i+thomson._palette.offset-1, 
				 thomson.levels.pc[r+1], 
				 thomson.levels.pc[g+1], 
				 thomson.levels.pc[b+1])
	end
	for i=thomson._palette.offset+thomson._palette.max,255 do
		setcolor(i,0,0,0)
	end
	-- refresh screen content
	clearpicture(thomson._palette.offset + thomson.border())
	for y=0,thomson.h-1 do
		for x=0,thomson.w-1 do
			local p = thomson.point(x,y)
			if p<0 then p=-p-1 end
			thomson._putpixel(x,y,thomson._palette.offset + p)
		end
	end
	updatescreen()
end

-- bitmap 16 mode
function thomson.setBM16() 
	-- put a pixel onto real screen
	function thomson._putpixel(x,y,c)
		putpicturepixel(x*2+0,y,c)
		putpicturepixel(x*2+1,y,c)
	end
	-- put a pixel in thomson screen
	function thomson.pset(x,y,c)
		local bank = x%4<2 and thomson.ramA or thomson.ramB
		local offs = math.floor(x/4)+y*40+1
		if x%2==0 then
			bank[offs] = (bank[offs]%16)+c*16
		else
			bank[offs] = math.floor(bank[offs]/16)*16+c
		end
		-- c=c+thomson._palette.offset
		-- putpicturepixel(x*2+0,y,c)
		-- putpicturepixel(x*2+1,y,c)
	end
	-- get thomson pixel at (x,y)
	function thomson.point(x,y)
		local bank = x%4<2 and thomson.ramA or thomson.ramB
		local offs = math.floor(x/4)+y*40+1
		if x%2==0 then
			return math.floor(bank[offs]/16)
		else
			return bank[offs]%16
		end
	end
	-- return internal MAP file
	function thomson._get_map_data()
		local tmp = {}
		for x=1,40 do
			for y=x,x+7960,40 do
				table.insert(tmp, thomson.ramA[y])
			end
			for y=x,x+7960,40 do
				table.insert(tmp, thomson.ramB[y])
			end
		end
		local pal = {}
		for i=1,16 do 
			pal[2*i-1] = math.floor(thomson._palette[i]/256)
			pal[2*i+0] =            thomson._palette[i]%256
		end
		-- build data
		local data={
			-- BM16
			0x40,
			-- ncols-1
			79,
			-- nlines-1
			24
		};
		thomson._compress(data, tmp)
		thomson._append(data,{0,0})
		-- padd to word
		if #data%2==1 then table.insert(data,0); end
		-- tosnap
		thomson._append(data,{0,128,0,thomson.border(),0,3})
		thomson._append(data, pal)
		thomson._append(data,{0xa5,0x5a})
		return data
	end
	
	thomson.w = 160
	thomson.h = 200
	thomson.palette(0,thomson.default_palette)
	thomson.border(0)
end

end -- thomson